@@ -14,15 +14,18 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import contextlib
 import inspect
 import multiprocessing
 import sys
 import threading
+from time import monotonic
 import traceback
 
-from monotonic import monotonic as now  # noqa
 import six
 
+now = monotonic
+
 try:
     import eventlet as _eventlet  # noqa
     EVENTLET_AVAILABLE = True
@@ -31,6 +34,8 @@ except ImportError:
 
 
 class WorkItem(object):
+    """A thing to be executed by a executor."""
+
     def __init__(self, future, fn, args, kwargs):
         self.future = future
         self.fn = fn
@@ -42,18 +47,27 @@ class WorkItem(object):
             return
         try:
             result = self.fn(*self.args, **self.kwargs)
-        except BaseException:
-            exc_type, exc_value, exc_tb = sys.exc_info()
+        except SystemExit as e:
             try:
-                if six.PY2:
-                    self.future.set_exception_info(exc_value, exc_tb)
-                else:
-                    self.future.set_exception(exc_value)
+                self.fail()
             finally:
-                del(exc_type, exc_value, exc_tb)
+                raise e
+        except BaseException:
+            self.fail()
         else:
             self.future.set_result(result)
 
+    def fail(self, exc_info=None):
+        exc_type, exc_value, exc_tb = exc_info or sys.exc_info()
+        try:
+            if six.PY2:
+                self.future.set_exception_info(exc_value, exc_tb)
+            else:
+                self.future.set_exception(exc_value)
+        finally:
+            if exc_info is None:
+                del exc_type, exc_value, exc_tb
+
 
 class Failure(object):
     """Object that captures a exception (and its associated information)."""
@@ -74,6 +88,14 @@ class Failure(object):
         finally:
             del exc_info
 
+    @property
+    def exc_type(self):
+        return self.exc_info[0]
+
+    @property
+    def exc_value(self):
+        return self.exc_info[1]
+
 
 def get_callback_name(cb):
     """Tries to get a callbacks fully-qualified name.
@@ -108,14 +130,19 @@ def get_callback_name(cb):
         return ".".join(segments)
 
 
-def get_optimal_thread_count(default=2):
+def get_optimal_thread_count(default=5):
     """Try to guess optimal thread count for current system."""
     try:
-        return multiprocessing.cpu_count() + 1
+        return multiprocessing.cpu_count() * 5
+    except NotImplementedError:
+        return default
+
+
+def get_optimal_process_count(default=1):
+    """Try to guess optimal process count for current system."""
+    try:
+        return multiprocessing.cpu_count()
     except NotImplementedError:
-        # NOTE(harlowja): apparently may raise so in this case we will
-        # just setup two threads since it's hard to know what else we
-        # should do in this situation.
         return default
 
 
@@ -126,11 +153,24 @@ class Barrier(object):
         self._active = 0
         self._cond = cond_cls()
 
+    @property
+    def active(self):
+        return self._active
+
     def incr(self):
         with self._cond:
             self._active += 1
             self._cond.notify_all()
 
+    @contextlib.contextmanager
+    def decr_cm(self):
+        with self._cond:
+            self._active -= 1
+            try:
+                yield self._active
+            finally:
+                self._cond.notify_all()
+
     def decr(self):
         with self._cond:
             self._active -= 1
